"""
plot_compare.py

This script plots 1+ timeseries figures of simulation data with overlain reference data.

Reference data channels without provincial stratification should either NOT have 'Province' or should
have 'All' as their 'Province value.
"""


import itertools
import json
import math
import matplotlib.pyplot as plt
import os
import pandas as pd

from matplotlib import collections as mc

from dtk.utils.observations.utils import parse_ingest_data_from_xlsm
from dtk.utils.observations.DataFrameWrapper import DataFrameWrapper
from dtk.utils.observations.AgeBin import AgeBin
from simtools.Analysis.AnalyzeManager import AnalyzeManager
from simtools.Analysis.BaseAnalyzers import DownloadAnalyzer
from simtools.SetupParser import SetupParser
from simtools.Utilities.Experiments import retrieve_experiment

from hiv.lib.utils.analysis import model_population_in_year

NO_PROVINCE = 'All'

# if a channel name has one of these in it, do not scale sim data by population ratio
NON_SCALING = ['Prevalence', 'Incidence']


def read_sim_data(fn, node_map):
    sim_df = pd.read_csv(fn, skipinitialspace=True)

    # fix up Node-Province column
    if 'Node' in list(sim_df.columns):
        sim_df = sim_df.set_index('Node').rename(node_map).reset_index().rename(columns={'Node': 'Province'})
    return sim_df


def download_files(filenames, output_directory, exp_id=None, sim_ids=None):
    if not ((exp_id is None) ^ (sim_ids is None)):
        raise Exception('Can either download an experiment or a list of sims, not both')

    analyzer = DownloadAnalyzer(filenames=filenames, output_path=output_directory)
    if exp_id:
        AnalyzeManager(exp_list=[exp_id], analyzers=analyzer).analyze()
    else:
        AnalyzeManager(sim_liist=sim_ids, analyzers=analyzer).analyze()


def make_collection(d):
    try:
        zipped_data = zip(d['Year'], d['Result'])
    except TypeError:
        zipped_data = zip([d['Year']], [d['Result']])
    return zipped_data


def scale_to_census(df, pop_df, census_population, census_year, census_min_age, census_max_age, verbose=False):
    model_pop, census_to_model_ratio = model_population_in_year(year=census_year, obs_population=census_population,
                                                                df=pop_df,
                                                                low_age=census_min_age, high_age=census_max_age,
                                                                population_col='Result', verbose=verbose)

    df = df.assign(Result=df['Result'] * census_to_model_ratio)
    return df


def is_scaling_channel(channel):
    do_scaling = True
    for pattern in NON_SCALING:
        if pattern in channel:
            do_scaling = False
            break
    return do_scaling


def generate_plot(reference, sim_filenames, pop_filenames, node_map, channel, census_population, census_year,
                  census_min_age, census_max_age, verbose=False):
    # one-off fix for Edinah's South Africa calibration
    # NATIONAL_PROVINCE = 'NationwideSA'
    # df = reference._dataframe
    # df = df.rename(columns={NATIONAL_PROVINCE: NO_PROVINCE})
    # reference = PopulationObs(dataframe=df, stratifiers=reference.stratifiers)

    try:
        provinces = reference.get_provinces()
        multi_index = ['AgeBin', 'Province', 'Gender']
        for province in provinces:
            if isinstance(province, str):
                continue
            if math.isnan(province):
                raise DataFrameWrapper.MissingRequiredData('')
        include_province = True
    except DataFrameWrapper.MissingRequiredData as e:
        provinces = [NO_PROVINCE]
        multi_index = ['AgeBin', 'Gender']
        include_province = False

    allowed_genders = ['Male', 'Female']
    genders = [gender for gender in reference.get_genders() if gender in allowed_genders]
    agebins = reference.get_age_bins()
    years = reference.get_years()

    # the rest of this method expects reference to be a multi-indexed DataFrame
    reference = reference._dataframe
    reference.set_index(multi_index, inplace=True)

    # population scaling for sim data, T/F
    do_scaling = is_scaling_channel(channel)
    if do_scaling:
        age_bin = AgeBin(start=census_min_age, end=census_max_age)
        print('-- Will perform scaling of sim data to census population, age bin: %s.' % age_bin)
    else:
        print('-- No population scaling will be performed (channel type is one of: %s)' % NON_SCALING)

    # pre-generate all figures to make
    figure_dict = {}
    for agebin in agebins:
        figure, axis = plt.subplots(len(genders), max(len(provinces), 2), figsize=(10, 6), sharex=True, sharey=True)
        figure_dict[agebin] = {'figure': figure, 'axis': axis}

    # read in data from simulation files
    sim_multi_index = ['AgeBin', 'Province', 'Gender']
    n_sims = len(sim_filenames)
    n = 1
    for i in range(len(sim_filenames)):
        fn = sim_filenames[i]

        if verbose:
            print('Processing sim file %d/%d ...' % (n, n_sims))
        results = read_sim_data(fn, node_map=node_map)
        results = results.sort_values(by='Year', ascending=True)

        # scale Results to census data
        if do_scaling:
            pop_filename = pop_filenames[i]
            pop = read_sim_data(pop_filename, node_map=node_map)
            results = scale_to_census(results, pop, census_population, census_year, census_min_age, census_max_age,
                                      verbose=verbose)
        results.set_index(sim_multi_index, inplace=True)

        # loop over all data tuples to plot, grabbing sim/reference data and applying to appropriate figure
        for agebin, province, gender in itertools.product(agebins, provinces, genders):
            ax = figure_dict[agebin]['axis']
            sim_ts_index = (agebin, province, gender)
            reference_ts_index = sim_ts_index if include_province else (agebin, gender)


            if sim_ts_index not in results.index:
                print('Reference: agebin %s province %s gender %s not in simulation output, skipping...' % (agebin, province, gender))
                continue

            gi = genders.index(gender)
            pi = provinces.index(province)

            # get simulation data and plot it
            sim_data = results.loc[[sim_ts_index]]
            if len(sim_data.index) == 1:
                # a single data element
                ax[gi, pi].plot(sim_data['Year'], sim_data['Result'], marker='.', color='b', alpha=0.1, ms=15)
            else:
                # a line of data
                coll = list(make_collection(sim_data))
                lc = mc.LineCollection([coll], linewidths=0.5)
                lc.set_alpha(0.1)
                lc.set_color('b')
                ax[gi, pi].add_collection(lc)

            # ger reference data and plot it
            ref_s = reference.loc[reference_ts_index]
            ax[gi, pi].plot(ref_s['Year'], ref_s[channel], marker='.', color='k', ms=15)

            # make the plot pretty
            ax[gi, pi].autoscale()
            ax[gi, pi].margins(x=0.1, y=0.1)

            ax[gi, pi].set_xticks(years)
            ax[gi, pi].set_xticklabels(years, rotation=75)
            if pi == 0:
                ax[gi, pi].set_ylabel('%s %s' % (gender, channel))

            if gi == 0:
                ax[gi, pi].set_title('%s' % province.replace(' ', '\n'), fontsize=8)

            if gi == len(genders) - 1:
                ax[gi, pi].set_xlabel('Year')
        n += 1

    # finalize figures
    for agebin, fd in figure_dict.items():
        fd['figure'].suptitle('%s %s' % (channel, agebin), fontsize=12)
        plt.subplots_adjust(top=0.9)
        fd['figure'].savefig('%s_%s.png' % (channel, agebin))


def main(args):
    # load the node name map; convert json str keys to int
    with open(args.node_map_file, 'r') as f:
        node_map = {int(k): v for k, v in json.load(f).items()}

    # get reference data

    # params is a dict, reference is a PopulationObs object, analyzers is a dict of AnalyzerClass: WEIGHT
    params, reference, analyzers = parse_ingest_data_from_xlsm(filename=args.reference_file)
    reference.fix_age_bins()  # currently necessary, converts from [0:5) -> [0, 5) AgeBin format
    reference = reference.filter(keep_only=[args.channel])

    # get sim data

    # obtain output data file to get plotting data from
    comps_output_root = os.path.join('output', 'post_process')

    # ck4, this is so ugly. We should require reference data to have the exact same name as our model/post-processed field
    if args.channel == 'On_ART':
        comps_download_field = 'Number_On_ART'
    else:
        comps_download_field = args.channel
    sim_filename = '%s.csv' % comps_download_field
    comps_filenames = [os.path.join(comps_output_root, sim_filename)]

    # add in Population for computing the pop scaling factor
    if is_scaling_channel(args.channel):
        comps_filenames.append(os.path.join(comps_output_root, 'Population.csv'))

    # ck4, hack to get 'top likelihood' sims plotted instead of random ones
    if args.plot_most_likely:
        calib_df = pd.read_csv('Selected_Trajectory_Params_Likelihoods_and_Metadata.csv').iloc[0:args.n_sims, :]
        simulation_ids = list(calib_df['simulation_id'])

        if args.download:
            download_files(filenames=comps_filenames, output_directory=args.output_directory, sim_ids=simulation_ids)

    else:
        if args.download:
            download_files(filenames=comps_filenames, output_directory=args.output_directory, exp_id=args.exp_id)

        # retrieve sim ids for this experiment
        experiment = retrieve_experiment(exp_id=args.exp_id)
        simulation_ids = []
        for i, sim in enumerate(experiment.simulations):
            if i >= args.n_sims:
                break
            simulation_ids.append(sim.id)


    # create requested plots
    # Make sure these two lists are in the same sim order
    result_filenames = [os.path.join(args.output_directory, sim_id, sim_filename) for sim_id in simulation_ids]
    population_filenames = [os.path.join(args.output_directory, sim_id, 'Population.csv') for sim_id in simulation_ids]

    generate_plot(reference, result_filenames, population_filenames, node_map=node_map, channel=args.channel,
                  census_population=args.census_population, census_year=args.census_year,
                  census_min_age=args.census_min_age, census_max_age=args.census_max_age, verbose=args.verbose)
    plt.show()


def parse_args():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--channel', dest='channel', type=str, required=True,
                        help='Channel to plot (Required).')
    parser.add_argument('-r', '--reference', dest='reference_file', type=str, required=True,
                        help='Reference xls data file to use (Required).')
    parser.add_argument('-id', dest='exp_id', type=str, required=True,
                        help='Experiment id from which to gather simulation data for plotting (Required).')
    parser.add_argument('-N', '--nodes', dest='node_map_file', type=str, required=True,
                        help='JSON file with node mapping information (Required).')
    parser.add_argument('-d', '--download', dest='download', action='store_true',
                        help='Download experiment/simulation data before plotting (Default: Use previously downloaded files).')
    parser.add_argument('-o', '--output_directory', dest='output_directory', default='output',
                        help='Directory to read/write simulation data from (Default: output).')
    parser.add_argument('-n', '--nsims', dest='n_sims', type=int, default=100,
                        help='Plot this number of simulations in the experiment (Default: (up to the) first 100).')
    parser.add_argument('-y', '--census_year', dest='census_year', type=int, required=True,
                        help='Year of census data (Integer year).')
    parser.add_argument('-p', '--population', dest='census_population', type=int, required=True,
                        help='Census population (Integer)')
    parser.add_argument('--minage', dest='census_min_age', type=int, required=True,
                        help='Census population minimum age (inclusive) (Integer) (Required).')
    parser.add_argument('--maxage', dest='census_max_age', type=int, required=True,
                        help='Census population maximum age (exclusive) (Integer) (Required).')
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help='Print more info to the screen during processing.')
    parser.add_argument('-P', '--plot_most_likely', dest='plot_most_likely', action='store_true',
                        help='Plot the most likely sims, not arbitrary ones.')

    args = parser.parse_args()
    return args


if __name__ == '__main__':
    args = parse_args()
    SetupParser.init(selected_block='HPC')
    main(args)
